import sys

import pytest

from typing_inspect import (
    is_generic_type, is_callable_type, is_new_type, is_tuple_type, is_union_type,
    is_optional_type, is_final_type, is_literal_type, is_typevar, is_classvar,
    is_forward_ref, get_origin, get_parameters, get_last_args, get_args, get_bound,
    get_constraints, get_generic_type, get_generic_bases, get_last_origin,
    typed_dict_keys, get_forward_arg, WITH_FINAL, WITH_LITERAL, LEGACY_TYPING, WITH_NEWTYPE,
)
from unittest import TestCase, main, skipIf, skipUnless
from typing import (
    Union, Callable, Optional, TypeVar, Sequence, AnyStr, Mapping,
    MutableMapping, Iterable, Generic, List, Any, Dict, Tuple, NamedTuple,
)

from mypy_extensions import TypedDict as METypedDict
from typing_extensions import TypedDict as TETypedDict
from typing_extensions import Final
from typing_extensions import Literal
from typing_extensions import NewType as NewType_

# Does this raise an exception ?
#      from typing import ClassVar
if sys.version_info < (3, 5, 3):
    WITH_CLASSVAR = False
    CLASSVAR_GENERIC = []
    CLASSVAR_TYPEVAR = []
else:
    from typing import ClassVar
    WITH_CLASSVAR = True
    CLASSVAR_GENERIC = [ClassVar[List[int]], ClassVar]
    CLASSVAR_TYPEVAR = [ClassVar[int]]


# Does this raise an exception ?
#     class Foo(Callable[[int], int]):
#         pass
if sys.version_info < (3, 5, 3):
    SUBCLASSABLE_CALLABLES = False
else:
    SUBCLASSABLE_CALLABLES = True


# Does this raise an exception ?
#     class MyClass(Tuple[str, int]):
#         pass
if sys.version_info < (3, 5, 3):
    SUBCLASSABLE_TUPLES = False
else:
    SUBCLASSABLE_TUPLES = True


# Does this raise an exception ?
#     T = TypeVar('T')
#     Union[T, str][int]
if sys.version_info < (3, 5, 3):
    EXISTING_UNIONS_SUBSCRIPTABLE = False
else:
    EXISTING_UNIONS_SUBSCRIPTABLE = True


# Does this raise an exception ?
#     Union[callable, Callable[..., int]]
if sys.version_info[:3] == (3, 5, 3) or sys.version_info[:3] < (3, 5, 2):
    UNION_SUPPORTS_BUILTIN_CALLABLE = False
else:
    UNION_SUPPORTS_BUILTIN_CALLABLE = True


# Does this raise an exception ?
#   Tuple[T][int]
#   List[Tuple[T]][int]
if sys.version_info[:3] == (3, 5, 3) or sys.version_info[:3] < (3, 5, 2):
    GENERIC_TUPLE_PARAMETRIZABLE = False
else:
    GENERIC_TUPLE_PARAMETRIZABLE = True


# Does this raise an exception ?
#    Dict[T, T][int]
#    Dict[int, Tuple[T, T]][int]
if sys.version_info[:3] == (3, 5, 3) or sys.version_info[:3] < (3, 5, 2):
    GENERIC_WITH_MULTIPLE_IDENTICAL_PARAMETERS_CAN_BE_PARAMETRIZED_ONCE = False
else:
    GENERIC_WITH_MULTIPLE_IDENTICAL_PARAMETERS_CAN_BE_PARAMETRIZED_ONCE = True


# Does this raise an exception ?
#    Callable[[T], int][int]
#    Callable[[], T][int]
if sys.version_info[:3] < (3, 5, 4):
    CALLABLE_CAN_BE_PARAMETRIZED = False
else:
    CALLABLE_CAN_BE_PARAMETRIZED = True


NEW_TYPING = sys.version_info[:3] >= (3, 7, 0)  # PEP 560

PY36_TESTS = """
class TDM(METypedDict):
    x: int
    y: int
class TDE(TETypedDict):
    x: int
    y: int
class Other(dict):
    x: int
    y: int
"""

PY36 = sys.version_info[:3] >= (3, 6, 0)
PY39 = sys.version_info[:3] >= (3, 9, 0)
if PY36:
    exec(PY36_TESTS)


# It is important for the test that this function is called 'NewType' to simulate the same __qualname__
# - which is "NewType.<locals>.new_type" - as typing.NewType has, i.e. it should be checked that is_new_type
# still do not accept a function which has the same __qualname__ and an attribute called __supertype__.
def NewType(name, tp):
    def new_type(x):
        return x

    new_type.__name__ = name
    new_type.__supertype__ = tp
    return new_type


class IsUtilityTestCase(TestCase):
    def sample_test(self, fun, samples, nonsamples):
        msg = "Error asserting that %s(%s) is %s"
        for s in samples:
            self.assertTrue(fun(s), msg=msg % (fun.__name__, str(s), 'True'))
        for s in nonsamples:
            self.assertFalse(fun(s), msg=msg % (fun.__name__, str(s), 'False'))

    def test_generic(self):
        T = TypeVar('T')
        samples = [Generic, Generic[T], Iterable[int], Mapping,
                   MutableMapping[T, List[int]], Sequence[Union[str, bytes]]]
        if PY39:
            samples.extend([list[int], dict[str, list[int]]])
        nonsamples = [int, Union[int, str], Union[int, T], Callable[..., T],
                      Optional, bytes, list] + CLASSVAR_GENERIC
        self.sample_test(is_generic_type, samples, nonsamples)

    def test_callable(self):
        samples = [Callable, Callable[..., int],
                   Callable[[int, int], Iterable[str]]]
        nonsamples = [int, type, 42, [], List[int]]
        if UNION_SUPPORTS_BUILTIN_CALLABLE:
            nonsamples.append(Union[callable, Callable[..., int]])
        self.sample_test(is_callable_type, samples, nonsamples)
        if SUBCLASSABLE_CALLABLES:
            class MyClass(Callable[[int], int]):
                pass
            self.assertTrue(is_callable_type(MyClass))

    def test_tuple(self):
        samples = [Tuple, Tuple[str, int], Tuple[Iterable, ...]]
        if PY39:
            samples.append(tuple[int, str])
        nonsamples = [int, tuple, 42, List[int], NamedTuple('N', [('x', int)])]
        self.sample_test(is_tuple_type, samples, nonsamples)
        if SUBCLASSABLE_TUPLES:
            class MyClass(Tuple[str, int]):
                pass
            self.assertTrue(is_tuple_type(MyClass))

    def test_union(self):
        T = TypeVar('T')
        S = TypeVar('S')
        samples = [Union, Union[T, int], Union[int, Union[T, S]]]
        nonsamples = [int, Union[int, int], [], Iterable[Any]]
        self.sample_test(is_union_type, samples, nonsamples)

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="requires 3.10 or higher")
    def test_union_pep604(self):
        T = TypeVar('T')
        S = TypeVar('S')
        samples = [T | int, int | (T | S), int | str]
        nonsamples = [int, int | int, [], Iterable[Any]]
        self.sample_test(is_union_type, samples, nonsamples)

    def test_optional_type(self):
        T = TypeVar('T')
        samples = [type(None),                # none type
                   Optional[int],             # direct union to none type 1
                   Optional[T],               # direct union to none type 2
                   Union[int, type(None)],    # direct union to none type 4
                   ]
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            samples += [Optional[T][int],          # direct union to none type 3
                        Union[str, T][type(None)]  # direct union to none type 5
                        ]

        # nested unions are supported
        samples += [Union[str, Optional[int]]]         # nested Union 1
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            samples += [Union[T, str][Optional[int]]]   # nested Union 2

        nonsamples = [int, Union[int, int], [], Iterable[Any], T]
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            nonsamples += [Union[T, str][int]]

        # unfortunately current definition sets these ones as non samples too
        S1 = TypeVar('S1', bound=Optional[int])
        S2 = TypeVar('S2', type(None), str)
        S3 = TypeVar('S3', Optional[int], str)
        S4 = TypeVar('S4', bound=Union[str, Optional[int]])
        nonsamples += [S1, S2, S3,                     # typevar bound or constrained to optional
                       Union[S1, int], S4              # combinations of the above
                       ]
        self.sample_test(is_optional_type, samples, nonsamples)

    @skipIf(not WITH_FINAL, "Final is not available")
    def test_final_type(self):
        samples = [
            Final,
            Final[int],
        ]
        nonsamples = [
            "v",
            1,
            (1, 2, 3),
            int,
            str,
            Union["u", "v"],
        ]
        self.sample_test(is_final_type, samples, nonsamples)

    @skipIf(not WITH_LITERAL, "Literal is not available")
    def test_literal_type(self):
        samples = [
            Literal,
            Literal["v"],
            Literal[1, 2, 3],
        ]
        nonsamples = [
            "v",
            (1, 2, 3),
            int,
            str,
            Union["u", "v"],
        ]
        self.sample_test(is_literal_type, samples, nonsamples)

    def test_typevar(self):
        T = TypeVar('T')
        S_co = TypeVar('S_co', covariant=True)
        samples = [T, S_co]
        nonsamples = [int, Union[T, int], Union[T, S_co], type] + CLASSVAR_TYPEVAR
        self.sample_test(is_typevar, samples, nonsamples)

    @skipIf(not WITH_CLASSVAR, "ClassVar is not present")
    def test_classvar(self):
        T = TypeVar('T')
        samples = [ClassVar, ClassVar[int], ClassVar[List[T]]]
        nonsamples = [int, 42, Iterable, List[int], type, T]
        self.sample_test(is_classvar, samples, nonsamples)

    @skipIf(not WITH_NEWTYPE, "NewType is not present")
    def test_new_type(self):
        T = TypeVar('T')

        class WithAttrSuperTypeCls:
            __supertype__ = str

        class WithAttrSuperTypeObj:
            def __init__(self):
                self.__supertype__ = str

        samples = [
            NewType_,
            NewType_('A', int),
            NewType_('B', complex),
            NewType_('C', List[int]),
            NewType_('D', Union['p', 'y', 't', 'h', 'o', 'n']),
            NewType_('E', List[Dict[str, float]]),
            NewType_('F', NewType('F_', int)),
        ]
        nonsamples = [
            int,
            42,
            Iterable,
            List[int],
            Union["u", "v"],
            type,
            T,
            NewType,
            NewType('N', int),
            WithAttrSuperTypeCls,
            WithAttrSuperTypeCls(),
            WithAttrSuperTypeObj,
            WithAttrSuperTypeObj(),
        ]
        self.sample_test(is_new_type, samples, nonsamples)

    def test_is_forward_ref(self):
        samples = []
        nonsamples = []
        for tp in (
            Union["FowardReference", Dict[str, List[int]]],
            Union["FR", List["FR"]],
            Optional["Fref"],
            Union["fRef", int],
            Union["fR", AnyStr],
        ):
            fr, not_fr = get_args(tp)
            samples.append(fr)
            nonsamples.append(not_fr)
        self.sample_test(is_forward_ref, samples, nonsamples)


class GetUtilityTestCase(TestCase):

    @skipIf(NEW_TYPING, "Not supported in Python 3.7")
    def test_last_origin(self):
        T = TypeVar('T')
        self.assertEqual(get_last_origin(int), None)
        if WITH_CLASSVAR:
            self.assertEqual(get_last_origin(ClassVar[int]), None)
        self.assertEqual(get_last_origin(Generic[T]), Generic)
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            self.assertEqual(get_last_origin(Union[T, int][str]), Union[T, int])
        if GENERIC_TUPLE_PARAMETRIZABLE:
            tp = List[Tuple[T, T]][int]
            self.assertEqual(get_last_origin(tp), List[Tuple[T, T]])
        self.assertEqual(get_last_origin(List), List)

    def test_origin(self):
        T = TypeVar('T')
        self.assertEqual(get_origin(int), None)
        if WITH_CLASSVAR:
            self.assertEqual(get_origin(ClassVar[int]), None)
        self.assertEqual(get_origin(Generic), Generic)
        self.assertEqual(get_origin(Generic[T]), Generic)
        # Cannot use assertEqual on Py3.5.2.
        self.assertIs(get_origin(Literal[42]), Literal)
        if PY39:
            self.assertEqual(get_origin(list[int]), list)
        if GENERIC_TUPLE_PARAMETRIZABLE:
            tp = List[Tuple[T, T]][int]
            self.assertEqual(get_origin(tp), list if NEW_TYPING else List)

    def test_parameters(self):
        T = TypeVar('T')
        S_co = TypeVar('S_co', covariant=True)
        U = TypeVar('U')
        self.assertEqual(get_parameters(int), ())
        self.assertEqual(get_parameters(Generic), ())
        self.assertEqual(get_parameters(Union), ())
        if not LEGACY_TYPING:
            self.assertEqual(get_parameters(List[int]), ())
        else:
            # in 3.5.3 a list has no __args__ and instead they are used in __parameters__
            # in 3.5.1 the behaviour is normal again.
            pass
        self.assertEqual(get_parameters(Generic[T]), (T,))
        self.assertEqual(get_parameters(Tuple[List[T], List[S_co]]), (T, S_co))
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            self.assertEqual(get_parameters(Union[S_co, Tuple[T, T]][int, U]), (U,))
        self.assertEqual(get_parameters(Mapping[T, Tuple[S_co, T]]), (T, S_co))
        if PY39:
            self.assertEqual(get_parameters(dict[int, T]), (T,))

    @skipIf(NEW_TYPING, "Not supported in Python 3.7")
    def test_last_args(self):
        T = TypeVar('T')
        S = TypeVar('S')
        self.assertEqual(get_last_args(int), ())
        self.assertEqual(get_last_args(Union), ())
        if WITH_CLASSVAR:
            self.assertEqual(get_last_args(ClassVar[int]), (int,))
        self.assertEqual(get_last_args(Union[T, int]), (T, int))
        self.assertEqual(get_last_args(Union[str, int]), (str, int))
        self.assertEqual(get_last_args(Tuple[T, int]), (T, int))
        self.assertEqual(get_last_args(Tuple[str, int]), (str, int))
        self.assertEqual(get_last_args(Generic[T]), (T, ))
        if GENERIC_TUPLE_PARAMETRIZABLE:
            tp = Iterable[Tuple[T, S]][int, T]
            self.assertEqual(get_last_args(tp), (int, T))
        if LEGACY_TYPING:
            self.assertEqual(get_last_args(Callable[[T, S], int]), (T, S))
            self.assertEqual(get_last_args(Callable[[], int]), ())
        else:
            self.assertEqual(get_last_args(Callable[[T, S], int]), (T, S, int))
            self.assertEqual(get_last_args(Callable[[], int]), (int,))

    @skipIf(NEW_TYPING, "Not supported in Python 3.7")
    def test_args(self):
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            T = TypeVar('T')
            self.assertEqual(get_args(Union[int, Tuple[T, int]][str]),
                             (int, (Tuple, str, int)))
            self.assertEqual(get_args(Union[int, Union[T, int], str][int]),
                             (int, str))
        self.assertEqual(get_args(int), ())

    def test_args_evaluated(self):
        T = TypeVar('T')
        if EXISTING_UNIONS_SUBSCRIPTABLE:
            self.assertEqual(get_args(Union[int, Tuple[T, int]][str], evaluate=True),
                             (int, Tuple[str, int]))
        if GENERIC_WITH_MULTIPLE_IDENTICAL_PARAMETERS_CAN_BE_PARAMETRIZED_ONCE:
            tp = Dict[int, Tuple[T, T]][Optional[int]]
            self.assertEqual(get_args(tp, evaluate=True),
                             (int, Tuple[Optional[int], Optional[int]]))
        if CALLABLE_CAN_BE_PARAMETRIZED:
            tp = Callable[[], T][int]
            self.assertEqual(get_args(tp, evaluate=True), ([], int,))

        self.assertEqual(get_args(Union[int, Callable[[Tuple[T, ...]], str]], evaluate=True),
                         (int, Callable[[Tuple[T, ...]], str]))

        # ClassVar special-casing
        if WITH_CLASSVAR:
            self.assertEqual(get_args(ClassVar, evaluate=True), ())
            self.assertEqual(get_args(ClassVar[int], evaluate=True), (int,))

        # Final special-casing
        if WITH_FINAL:
            self.assertEqual(get_args(Final, evaluate=True), ())
            self.assertEqual(get_args(Final[int], evaluate=True), (int,))

        # Literal special-casing
        if WITH_LITERAL:
            self.assertEqual(get_args(Literal, evaluate=True), ())
            self.assertEqual(get_args(Literal["value"], evaluate=True), ("value",))
            self.assertEqual(get_args(Literal[1, 2, 3], evaluate=True), (1, 2, 3))

        if PY39:
            self.assertEqual(get_args(list[int]), (int,))
            self.assertEqual(get_args(tuple[int, str]), (int, str))
            self.assertEqual(get_args(list[list[int]]), (list[int],))
            # This would return (~T,) before Python 3.9.
            self.assertEqual(get_args(List), ())

        if sys.version_info >= (3, 10):
            self.assertEqual(get_args(int | str), (int, str))
            self.assertEqual(get_args((int | tuple[T, int])[str]), (int, tuple[str, int]))

    def test_bound(self):
        T = TypeVar('T')
        TB = TypeVar('TB', bound=int)
        self.assertEqual(get_bound(T), None)
        self.assertEqual(get_bound(TB), int)

    def test_constraints(self):
        T = TypeVar('T')
        TC = TypeVar('TC', int, str)
        self.assertEqual(get_constraints(T), ())
        self.assertEqual(get_constraints(TC), (int, str))

    def test_generic_type(self):
        T = TypeVar('T')
        class Node(Generic[T]): pass
        self.assertIs(get_generic_type(Node()), Node)
        if not LEGACY_TYPING:
            self.assertIs(get_generic_type(Node[int]()), Node[int])
            self.assertIs(get_generic_type(Node[T]()), Node[T],)
        else:
            # Node[int]() was creating an object of NEW type Node[~T]
            # and Node[T]() was creating an object of NEW type Node[~T]
            pass
        self.assertIs(get_generic_type(1), int)

    def test_generic_bases(self):
        class MyClass(List[int], Mapping[str, List[int]]): pass
        self.assertEqual(get_generic_bases(MyClass),
                         (List[int], Mapping[str, List[int]]))
        self.assertEqual(get_generic_bases(int), ())

    @skipUnless(PY36, "Python 3.6 required")
    def test_typed_dict_mypy_extension(self):
        TDOld = METypedDict("TDOld", {'x': int, 'y': int})
        self.assertEqual(typed_dict_keys(TDM), {'x': int, 'y': int})
        self.assertEqual(typed_dict_keys(TDOld), {'x': int, 'y': int})
        self.assertIs(typed_dict_keys(dict), None)
        self.assertIs(typed_dict_keys(Other), None)
        self.assertIsNot(typed_dict_keys(TDM), TDM.__annotations__)

    @skipUnless(PY36, "Python 3.6 required")
    def test_typed_dict_typing_extension(self):
        TDOld = TETypedDict("TDOld", {'x': int, 'y': int})
        self.assertEqual(typed_dict_keys(TDE), {'x': int, 'y': int})
        self.assertEqual(typed_dict_keys(TDOld), {'x': int, 'y': int})
        self.assertIs(typed_dict_keys(dict), None)
        self.assertIs(typed_dict_keys(Other), None)
        self.assertIsNot(typed_dict_keys(TDE), TDE.__annotations__)

    @skipIf(
        (3, 5, 2) > sys.version_info[:3] >= (3, 5, 0),
        "get_args doesn't work in Python 3.5.0 and 3.5.1 for type"
        " List and ForwardRef arg"
    )
    def test_get_forward_arg(self):
        tp = List["FRef"]
        fr = get_args(tp)[0]
        self.assertEqual(get_forward_arg(fr), "FRef")
        self.assertEqual(get_forward_arg(tp), None)


if __name__ == '__main__':
    main()



"""Defines experimental API for runtime inspection of types defined
in the standard "typing" module.

Example usage::
    from typing_inspect import is_generic_type
"""

# NOTE: This module must support Python 2.7 in addition to Python 3.x

import sys
import types

from mypy_extensions import _TypedDictMeta as _TypedDictMeta_Mypy

# See comments in typing_extensions source on why the switch is at 3.9.2
if (3, 4, 0) <= sys.version_info[:3] < (3, 9, 2):
    from typing_extensions import _TypedDictMeta as _TypedDictMeta_TE
elif sys.version_info[:3] >= (3, 9, 2):
    # Situation with typing_extensions.TypedDict is complicated.
    # Use the one defined in typing_extentions, and if there is none,
    # fall back to typing.
    try:
        from typing_extensions import _TypedDictMeta as _TypedDictMeta_TE
    except ImportError:
        from typing import _TypedDictMeta as _TypedDictMeta_TE
else:
    # typing_extensions.TypedDict is a re-export from typing.
    from typing import TypedDict
    _TypedDictMeta_TE = type(TypedDict)

NEW_TYPING = sys.version_info[:3] >= (3, 7, 0)  # PEP 560
if NEW_TYPING:
    import collections.abc

WITH_FINAL = True
WITH_LITERAL = True
WITH_CLASSVAR = True
WITH_NEWTYPE = True
LEGACY_TYPING = False

if NEW_TYPING:
    from typing import (
        Generic, Callable, Union, TypeVar, ClassVar, Tuple, _GenericAlias,
        ForwardRef, NewType,
    )
    from typing_extensions import Final, Literal
    if sys.version_info[:3] >= (3, 9, 0):
        from typing import _SpecialGenericAlias
        typingGenericAlias = (_GenericAlias, _SpecialGenericAlias, types.GenericAlias)
    else:
        typingGenericAlias = (_GenericAlias,)
else:
    from typing import (
        Callable, CallableMeta, Union, Tuple, TupleMeta, TypeVar, GenericMeta,
        _ForwardRef,
    )
    try:
        from typing import _Union, _ClassVar
    except ImportError:
        # support for very old typing module <=3.5.3
        _Union = type(Union)
        WITH_CLASSVAR = False
        LEGACY_TYPING = True

    try:  # python 3.6
        from typing_extensions import _Final
    except ImportError:  # python 2.7
        try:
            from typing import _Final
        except ImportError:
            WITH_FINAL = False

    try:  # python 3.6
        from typing_extensions import Literal
    except ImportError:  # python 2.7
        try:
            from typing import Literal
        except ImportError:
            WITH_LITERAL = False

    try:  # python < 3.5.2
        from typing_extensions import NewType
    except ImportError:
        try:
            from typing import NewType
        except ImportError:
            WITH_NEWTYPE = False


def _gorg(cls):
    """This function exists for compatibility with old typing versions."""
    assert isinstance(cls, GenericMeta)
    if hasattr(cls, '_gorg'):
        return cls._gorg
    while cls.__origin__ is not None:
        cls = cls.__origin__
    return cls


def is_generic_type(tp):
    """Test if the given type is a generic type. This includes Generic itself, but
    excludes special typing constructs such as Union, Tuple, Callable, ClassVar.
    Examples::

        is_generic_type(int) == False
        is_generic_type(Union[int, str]) == False
        is_generic_type(Union[int, T]) == False
        is_generic_type(ClassVar[List[int]]) == False
        is_generic_type(Callable[..., T]) == False

        is_generic_type(Generic) == True
        is_generic_type(Generic[T]) == True
        is_generic_type(Iterable[int]) == True
        is_generic_type(Mapping) == True
        is_generic_type(MutableMapping[T, List[int]]) == True
        is_generic_type(Sequence[Union[str, bytes]]) == True
    """
    if NEW_TYPING:
        return (isinstance(tp, type) and issubclass(tp, Generic) or
                isinstance(tp, typingGenericAlias) and
                tp.__origin__ not in (Union, tuple, ClassVar, collections.abc.Callable))
    return (isinstance(tp, GenericMeta) and not
            isinstance(tp, (CallableMeta, TupleMeta)))


def is_callable_type(tp):
    """Test if the type is a generic callable type, including subclasses
    excluding non-generic types and callables.
    Examples::

        is_callable_type(int) == False
        is_callable_type(type) == False
        is_callable_type(Callable) == True
        is_callable_type(Callable[..., int]) == True
        is_callable_type(Callable[[int, int], Iterable[str]]) == True
        class MyClass(Callable[[int], int]):
            ...
        is_callable_type(MyClass) == True

    For more general tests use callable(), for more precise test
    (excluding subclasses) use::

        get_origin(tp) is collections.abc.Callable  # Callable prior to Python 3.7
    """
    if NEW_TYPING:
        return (tp is Callable or isinstance(tp, typingGenericAlias) and
                tp.__origin__ is collections.abc.Callable or
                isinstance(tp, type) and issubclass(tp, Generic) and
                issubclass(tp, collections.abc.Callable))
    return type(tp) is CallableMeta


def is_tuple_type(tp):
    """Test if the type is a generic tuple type, including subclasses excluding
    non-generic classes.
    Examples::

        is_tuple_type(int) == False
        is_tuple_type(tuple) == False
        is_tuple_type(Tuple) == True
        is_tuple_type(Tuple[str, int]) == True
        class MyClass(Tuple[str, int]):
            ...
        is_tuple_type(MyClass) == True

    For more general tests use issubclass(..., tuple), for more precise test
    (excluding subclasses) use::

        get_origin(tp) is tuple  # Tuple prior to Python 3.7
    """
    if NEW_TYPING:
        return (tp is Tuple or isinstance(tp, typingGenericAlias) and
                tp.__origin__ is tuple or
                isinstance(tp, type) and issubclass(tp, Generic) and
                issubclass(tp, tuple))
    return type(tp) is TupleMeta


def is_optional_type(tp):
    """Test if the type is type(None), or is a direct union with it, such as Optional[T].

    NOTE: this method inspects nested `Union` arguments but not `TypeVar` definition
    bounds and constraints. So it will return `False` if
     - `tp` is a `TypeVar` bound, or constrained to, an optional type
     - `tp` is a `Union` to a `TypeVar` bound or constrained to an optional type,
     - `tp` refers to a *nested* `Union` containing an optional type or one of the above.

    Users wishing to check for optionality in types relying on type variables might wish
    to use this method in combination with `get_constraints` and `get_bound`
    """

    if tp is type(None):  # noqa
        return True
    elif is_union_type(tp):
        return any(is_optional_type(tt) for tt in get_args(tp, evaluate=True))
    else:
        return False


def is_final_type(tp):
    """Test if the type is a final type. Examples::

        is_final_type(int) == False
        is_final_type(Final) == True
        is_final_type(Final[int]) == True
    """
    if NEW_TYPING:
        return (tp is Final or
                isinstance(tp, typingGenericAlias) and tp.__origin__ is Final)
    return WITH_FINAL and type(tp) is _Final


try:
    MaybeUnionType = types.UnionType
except AttributeError:
    MaybeUnionType = None


def is_union_type(tp):
    """Test if the type is a union type. Examples::

        is_union_type(int) == False
        is_union_type(Union) == True
        is_union_type(Union[int, int]) == False
        is_union_type(Union[T, int]) == True
        is_union_type(int | int) == False
        is_union_type(T | int) == True
    """
    if NEW_TYPING:
        return (tp is Union or
                (isinstance(tp, typingGenericAlias) and tp.__origin__ is Union) or
                (MaybeUnionType and isinstance(tp, MaybeUnionType)))
    return type(tp) is _Union


def is_literal_type(tp):
    if NEW_TYPING:
        return (tp is Literal or
                isinstance(tp, typingGenericAlias) and tp.__origin__ is Literal)
    return WITH_LITERAL and type(tp) is type(Literal)


def is_typevar(tp):
    """Test if the type represents a type variable. Examples::

        is_typevar(int) == False
        is_typevar(T) == True
        is_typevar(Union[T, int]) == False
    """

    return type(tp) is TypeVar


def is_classvar(tp):
    """Test if the type represents a class variable. Examples::

        is_classvar(int) == False
        is_classvar(ClassVar) == True
        is_classvar(ClassVar[int]) == True
        is_classvar(ClassVar[List[T]]) == True
    """
    if NEW_TYPING:
        return (tp is ClassVar or
                isinstance(tp, typingGenericAlias) and tp.__origin__ is ClassVar)
    elif WITH_CLASSVAR:
        return type(tp) is _ClassVar
    else:
        return False


def is_new_type(tp):
    """Tests if the type represents a distinct type. Examples::

        is_new_type(int) == False
        is_new_type(NewType) == True
        is_new_type(NewType('Age', int)) == True
        is_new_type(NewType('Scores', List[Dict[str, float]])) == True
    """
    if not WITH_NEWTYPE:
        return False
    elif sys.version_info[:3] >= (3, 10, 0) and sys.version_info.releaselevel != 'beta':
        return tp is NewType or isinstance(tp, NewType)
    elif sys.version_info[:3] >= (3, 0, 0):
        return (tp is NewType or
                (getattr(tp, '__supertype__', None) is not None and
                 getattr(tp, '__qualname__', '') == 'NewType.<locals>.new_type' and
                 tp.__module__ in ('typing', 'typing_extensions')))
    else:  # python 2
        # __qualname__ is not available in python 2, so we simplify the test here
        return (tp is NewType or
                (getattr(tp, '__supertype__', None) is not None and
                 tp.__module__ in ('typing', 'typing_extensions')))


def is_forward_ref(tp):
    """Tests if the type is a :class:`typing.ForwardRef`. Examples::

        u = Union["Milk", Way]
        args = get_args(u)
        is_forward_ref(args[0]) == True
        is_forward_ref(args[1]) == False
    """
    if not NEW_TYPING:
        return isinstance(tp, _ForwardRef)
    return isinstance(tp, ForwardRef)


def get_last_origin(tp):
    """Get the last base of (multiply) subscripted type. Supports generic types,
    Union, Callable, and Tuple. Returns None for unsupported types.
    Examples::

        get_last_origin(int) == None
        get_last_origin(ClassVar[int]) == None
        get_last_origin(Generic[T]) == Generic
        get_last_origin(Union[T, int][str]) == Union[T, int]
        get_last_origin(List[Tuple[T, T]][int]) == List[Tuple[T, T]]
        get_last_origin(List) == List
    """
    if NEW_TYPING:
        raise ValueError('This function is only supported in Python 3.6,'
                         ' use get_origin instead')
    sentinel = object()
    origin = getattr(tp, '__origin__', sentinel)
    if origin is sentinel:
        return None
    if origin is None:
        return tp
    return origin


def get_origin(tp):
    """Get the unsubscripted version of a type. Supports generic types, Union,
    Callable, and Tuple. Returns None for unsupported types. Examples::

        get_origin(int) == None
        get_origin(ClassVar[int]) == None
        get_origin(Generic) == Generic
        get_origin(Generic[T]) == Generic
        get_origin(Union[T, int]) == Union
        get_origin(List[Tuple[T, T]][int]) == list  # List prior to Python 3.7
    """
    if NEW_TYPING:
        if isinstance(tp, typingGenericAlias):
            return tp.__origin__ if tp.__origin__ is not ClassVar else None
        if tp is Generic:
            return Generic
        return None
    if isinstance(tp, GenericMeta):
        return _gorg(tp)
    if is_union_type(tp):
        return Union
    if is_tuple_type(tp):
        return Tuple
    if is_literal_type(tp):
        return Literal

    return None


def get_parameters(tp):
    """Return type parameters of a parameterizable type as a tuple
    in lexicographic order. Parameterizable types are generic types,
    unions, tuple types and callable types. Examples::

        get_parameters(int) == ()
        get_parameters(Generic) == ()
        get_parameters(Union) == ()
        get_parameters(List[int]) == ()

        get_parameters(Generic[T]) == (T,)
        get_parameters(Tuple[List[T], List[S_co]]) == (T, S_co)
        get_parameters(Union[S_co, Tuple[T, T]][int, U]) == (U,)
        get_parameters(Mapping[T, Tuple[S_co, T]]) == (T, S_co)
    """
    if LEGACY_TYPING:
        # python <= 3.5.2
        if is_union_type(tp):
            params = []
            for arg in (tp.__union_params__ if tp.__union_params__ is not None else ()):
                params += get_parameters(arg)
            return tuple(params)
        elif is_tuple_type(tp):
            params = []
            for arg in (tp.__tuple_params__ if tp.__tuple_params__ is not None else ()):
                params += get_parameters(arg)
            return tuple(params)
        elif is_generic_type(tp):
            params = []
            base_params = tp.__parameters__
            if base_params is None:
                return ()
            for bp_ in base_params:
                for bp in (get_args(bp_) if is_tuple_type(bp_) else (bp_,)):
                    if _has_type_var(bp) and not isinstance(bp, TypeVar):
                        raise TypeError(
                            "Cannot inherit from a generic class "
                            "parameterized with "
                            "non-type-variable %s" % bp)
                    if params is None:
                        params = []
                    if bp not in params:
                        params.append(bp)
            if params is not None:
                return tuple(params)
            else:
                return ()
        else:
            return ()
    elif NEW_TYPING:
        if (isinstance(tp, typingGenericAlias) or
                isinstance(tp, type) and issubclass(tp, Generic) and
                tp is not Generic):
            return tp.__parameters__
        else:
            return ()
    elif (
        is_generic_type(tp) or is_union_type(tp) or
        is_callable_type(tp) or is_tuple_type(tp)
    ):
        return tp.__parameters__ if tp.__parameters__ is not None else ()
    else:
        return ()


def get_last_args(tp):
    """Get last arguments of (multiply) subscripted type.
       Parameters for Callable are flattened. Examples::

        get_last_args(int) == ()
        get_last_args(Union) == ()
        get_last_args(ClassVar[int]) == (int,)
        get_last_args(Union[T, int]) == (T, int)
        get_last_args(Iterable[Tuple[T, S]][int, T]) == (int, T)
        get_last_args(Callable[[T], int]) == (T, int)
        get_last_args(Callable[[], int]) == (int,)
    """
    if NEW_TYPING:
        raise ValueError('This function is only supported in Python 3.6,'
                         ' use get_args instead')
    elif is_classvar(tp):
        return (tp.__type__,) if tp.__type__ is not None else ()
    elif is_generic_type(tp):
        try:
            if tp.__args__ is not None and len(tp.__args__) > 0:
                return tp.__args__
        except AttributeError:
            # python 3.5.1
            pass
        return tp.__parameters__ if tp.__parameters__ is not None else ()
    elif is_union_type(tp):
        try:
            return tp.__args__ if tp.__args__ is not None else ()
        except AttributeError:
            # python 3.5.2
            return tp.__union_params__ if tp.__union_params__ is not None else ()
    elif is_callable_type(tp):
        return tp.__args__ if tp.__args__ is not None else ()
    elif is_tuple_type(tp):
        try:
            return tp.__args__ if tp.__args__ is not None else ()
        except AttributeError:
            # python 3.5.2
            return tp.__tuple_params__ if tp.__tuple_params__ is not None else ()
    else:
        return ()


def _eval_args(args):
    """Internal helper for get_args."""
    res = []
    for arg in args:
        if not isinstance(arg, tuple):
            res.append(arg)
        elif is_callable_type(arg[0]):
            callable_args = _eval_args(arg[1:])
            if len(arg) == 2:
                res.append(Callable[[], callable_args[0]])
            elif arg[1] is Ellipsis:
                res.append(Callable[..., callable_args[1]])
            else:
                res.append(Callable[list(callable_args[:-1]), callable_args[-1]])
        else:
            res.append(type(arg[0]).__getitem__(arg[0], _eval_args(arg[1:])))
    return tuple(res)


def get_args(tp, evaluate=None):
    """Get type arguments with all substitutions performed. For unions,
    basic simplifications used by Union constructor are performed.
    On versions prior to 3.7 if `evaluate` is False (default),
    report result as nested tuple, this matches
    the internal representation of types. If `evaluate` is True
    (or if Python version is 3.7 or greater), then all
    type parameters are applied (this could be time and memory expensive).
    Examples::

        get_args(int) == ()
        get_args(Union[int, Union[T, int], str][int]) == (int, str)
        get_args(Union[int, Tuple[T, int]][str]) == (int, (Tuple, str, int))

        get_args(Union[int, Tuple[T, int]][str], evaluate=True) == \
                 (int, Tuple[str, int])
        get_args(Dict[int, Tuple[T, T]][Optional[int]], evaluate=True) == \
                 (int, Tuple[Optional[int], Optional[int]])
        get_args(Callable[[], T][int], evaluate=True) == ([], int,)
    """
    if NEW_TYPING:
        if evaluate is not None and not evaluate:
            raise ValueError('evaluate can only be True in Python >= 3.7')
        # Note special aliases on Python 3.9 don't have __args__.
        if isinstance(tp, typingGenericAlias) and hasattr(tp, '__args__'):
            res = tp.__args__
            if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res
        if MaybeUnionType and isinstance(tp, MaybeUnionType):
            return tp.__args__
        return ()
    if is_classvar(tp) or is_final_type(tp):
        return (tp.__type__,) if tp.__type__ is not None else ()
    if is_literal_type(tp):
        return tp.__values__ or ()
    if (
        is_generic_type(tp) or is_union_type(tp) or
        is_callable_type(tp) or is_tuple_type(tp)
    ):
        try:
            tree = tp._subs_tree()
        except AttributeError:
            # Old python typing module <= 3.5.3
            if is_union_type(tp):
                # backport of union's subs_tree
                tree = _union_subs_tree(tp)
            elif is_generic_type(tp):
                # backport of GenericMeta's subs_tree
                tree = _generic_subs_tree(tp)
            elif is_tuple_type(tp):
                # ad-hoc (inspired by union)
                tree = _tuple_subs_tree(tp)
            else:
                # tree = _subs_tree(tp)
                return ()

        if isinstance(tree, tuple) and len(tree) > 1:
            if not evaluate:
                return tree[1:]
            res = _eval_args(tree[1:])
            if get_origin(tp) is Callable and res[0] is not Ellipsis:
                res = (list(res[:-1]), res[-1])
            return res

    return ()


def get_bound(tp):
    """Return the type bound to a `TypeVar` if any.

    It the type is not a `TypeVar`, a `TypeError` is raised.
    Examples::

        get_bound(TypeVar('T')) == None
        get_bound(TypeVar('T', bound=int)) == int
    """

    if is_typevar(tp):
        return getattr(tp, '__bound__', None)
    else:
        raise TypeError("type is not a `TypeVar`: " + str(tp))


def get_constraints(tp):
    """Returns the constraints of a `TypeVar` if any.

    It the type is not a `TypeVar`, a `TypeError` is raised
    Examples::

        get_constraints(TypeVar('T')) == ()
        get_constraints(TypeVar('T', int, str)) == (int, str)
    """

    if is_typevar(tp):
        return getattr(tp, '__constraints__', ())
    else:
        raise TypeError("type is not a `TypeVar`: " + str(tp))


def get_generic_type(obj):
    """Get the generic type of an object if possible, or runtime class otherwise.
    Examples::

        class Node(Generic[T]):
            ...
        type(Node[int]()) == Node
        get_generic_type(Node[int]()) == Node[int]
        get_generic_type(Node[T]()) == Node[T]
        get_generic_type(1) == int
    """

    gen_type = getattr(obj, '__orig_class__', None)
    return gen_type if gen_type is not None else type(obj)


def get_generic_bases(tp):
    """Get generic base types of a type or empty tuple if not possible.
    Example::

        class MyClass(List[int], Mapping[str, List[int]]):
            ...
        MyClass.__bases__ == (List, Mapping)
        get_generic_bases(MyClass) == (List[int], Mapping[str, List[int]])
    """
    if LEGACY_TYPING:
        return tuple(t for t in tp.__bases__ if isinstance(t, GenericMeta))
    else:
        return getattr(tp, '__orig_bases__', ())


def typed_dict_keys(td):
    """If td is a TypedDict class, return a dictionary mapping the typed keys to types.
    Otherwise, return None. Examples::

        class TD(TypedDict):
            x: int
            y: int
        class Other(dict):
            x: int
            y: int

        typed_dict_keys(TD) == {'x': int, 'y': int}
        typed_dict_keys(dict) == None
        typed_dict_keys(Other) == None
    """
    if isinstance(td, (_TypedDictMeta_Mypy, _TypedDictMeta_TE)):
        return td.__annotations__.copy()
    return None


def get_forward_arg(fr):
    """
    If fr is a ForwardRef, return the string representation of the forward reference.
    Otherwise return None. Examples::

        tp = List["FRef"]
        fr = get_args(tp)[0]
        get_forward_arg(fr) == "FRef"
        get_forward_arg(tp) == None
    """
    return fr.__forward_arg__ if is_forward_ref(fr) else None


# A few functions backported and adapted for the LEGACY_TYPING context, and used above

def _replace_arg(arg, tvars, args):
    """backport of _replace_arg"""
    if tvars is None:
        tvars = []
    # if hasattr(arg, '_subs_tree') and isinstance(arg, (GenericMeta, _TypingBase)):
    #     return arg._subs_tree(tvars, args)
    if is_union_type(arg):
        return _union_subs_tree(arg, tvars, args)
    if is_tuple_type(arg):
        return _tuple_subs_tree(arg, tvars, args)
    if is_generic_type(arg):
        return _generic_subs_tree(arg, tvars, args)
    if isinstance(arg, TypeVar):
        for i, tvar in enumerate(tvars):
            if arg == tvar:
                return args[i]
    return arg


def _remove_dups_flatten(parameters):
    """backport of _remove_dups_flatten"""

    # Flatten out Union[Union[...], ...].
    params = []
    for p in parameters:
        if isinstance(p, _Union):  # and p.__origin__ is Union:
            params.extend(p.__union_params__)  # p.__args__)
        elif isinstance(p, tuple) and len(p) > 0 and p[0] is Union:
            params.extend(p[1:])
        else:
            params.append(p)
    # Weed out strict duplicates, preserving the first of each occurrence.
    all_params = set(params)
    if len(all_params) < len(params):
        new_params = []
        for t in params:
            if t in all_params:
                new_params.append(t)
                all_params.remove(t)
        params = new_params
        assert not all_params, all_params
    # Weed out subclasses.
    # E.g. Union[int, Employee, Manager] == Union[int, Employee].
    # If object is present it will be sole survivor among proper classes.
    # Never discard type variables.
    # (In particular, Union[str, AnyStr] != AnyStr.)
    all_params = set(params)
    for t1 in params:
        if not isinstance(t1, type):
            continue
        if any(isinstance(t2, type) and issubclass(t1, t2)
               for t2 in all_params - {t1}
               if (not (isinstance(t2, GenericMeta) and
                        get_origin(t2) is not None) and
                   not isinstance(t2, TypeVar))):
            all_params.remove(t1)
    return tuple(t for t in params if t in all_params)


def _subs_tree(cls, tvars=None, args=None):
    """backport of typing._subs_tree, adapted for legacy versions """
    def _get_origin(cls):
        try:
            return cls.__origin__
        except AttributeError:
            return None

    current = _get_origin(cls)
    if current is None:
        if not is_union_type(cls) and not is_tuple_type(cls):
            return cls

    # Make of chain of origins (i.e. cls -> cls.__origin__)
    orig_chain = []
    while _get_origin(current) is not None:
        orig_chain.append(current)
        current = _get_origin(current)

    # Replace type variables in __args__ if asked ...
    tree_args = []

    def _get_args(cls):
        if is_union_type(cls):
            cls_args = cls.__union_params__
        elif is_tuple_type(cls):
            cls_args = cls.__tuple_params__
        else:
            try:
                cls_args = cls.__args__
            except AttributeError:
                cls_args = ()
        return cls_args if cls_args is not None else ()

    for arg in _get_args(cls):
        tree_args.append(_replace_arg(arg, tvars, args))
    # ... then continue replacing down the origin chain.
    for ocls in orig_chain:
        new_tree_args = []
        for arg in _get_args(ocls):
            new_tree_args.append(_replace_arg(arg, get_parameters(ocls), tree_args))
        tree_args = new_tree_args
    return tree_args


def _union_subs_tree(tp, tvars=None, args=None):
    """ backport of Union._subs_tree """
    if tp is Union:
        return Union  # Nothing to substitute
    tree_args = _subs_tree(tp, tvars, args)
    # tree_args = tp.__union_params__ if tp.__union_params__ is not None else ()
    tree_args = _remove_dups_flatten(tree_args)
    if len(tree_args) == 1:
        return tree_args[0]  # Union of a single type is that type
    return (Union,) + tree_args


def _generic_subs_tree(tp, tvars=None, args=None):
    """ backport of GenericMeta._subs_tree """
    if tp.__origin__ is None:
        return tp
    tree_args = _subs_tree(tp, tvars, args)
    return (_gorg(tp),) + tuple(tree_args)


def _tuple_subs_tree(tp, tvars=None, args=None):
    """ ad-hoc function (inspired by union) for legacy typing """
    if tp is Tuple:
        return Tuple  # Nothing to substitute
    tree_args = _subs_tree(tp, tvars, args)
    return (Tuple,) + tuple(tree_args)


def _has_type_var(t):
    if t is None:
        return False
    elif is_union_type(t):
        return _union_has_type_var(t)
    elif is_tuple_type(t):
        return _tuple_has_type_var(t)
    elif is_generic_type(t):
        return _generic_has_type_var(t)
    elif is_callable_type(t):
        return _callable_has_type_var(t)
    else:
        return False


def _union_has_type_var(tp):
    if tp.__union_params__:
        for t in tp.__union_params__:
            if _has_type_var(t):
                return True
    return False


def _tuple_has_type_var(tp):
    if tp.__tuple_params__:
        for t in tp.__tuple_params__:
            if _has_type_var(t):
                return True
    return False


def _callable_has_type_var(tp):
    if tp.__args__:
        for t in tp.__args__:
            if _has_type_var(t):
                return True
    return _has_type_var(tp.__result__)


def _generic_has_type_var(tp):
    if tp.__parameters__:
        for t in tp.__parameters__:
            if _has_type_var(t):
                return True
    return False



#!/usr/bin/env python

# NOTE: This package must support Python 2.7 in addition to Python 3.x

import sys
from setuptools import setup

version = '0.8.0'
description = 'Runtime inspection utilities for typing module.'
long_description = '''
Typing Inspect
==============

The "typing_inspect" module defines experimental API for runtime
inspection of types defined in the standard "typing" module.
'''.lstrip()

classifiers = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Software Development',
]

install_requires = [
    'mypy_extensions >= 0.3.0',
    'typing >= 3.7.4;python_version<"3.5"',
    'typing_extensions >= 3.7.4',
]

setup(
    name='typing_inspect',
    version=version,
    description=description,
    long_description=long_description,
    author='Ivan Levkivskyi',
    author_email='levkivskyi@gmail.com',
    url='https://github.com/ilevkivskyi/typing_inspect',
    license='MIT',
    keywords='typing function annotations type hints hinting checking '
             'checker typehints typehinting typechecking inspect '
             'reflection introspection',
    py_modules=['typing_inspect'],
    classifiers=classifiers,
    install_requires=install_requires,
)
